#include <vector>
#include <boost/noncopyable.hpp>

#ifndef CME_FAST_CURRENT_VALUES_H_
#define CME_FAST_CURRENT_VALUES_H_


#include "CValues.h"


using std::vector;

namespace FAST
{

/**
 * This class represents the current state of the decoded information.
 */
class CurrentValues : private boost::noncopyable
{
  private:
    const int _slots;
    vector<CInt32::type> i32_vals;
    vector<CInt64::type> i64_vals;
    vector<CuInt32::type> u32_vals;
    vector<CuInt64::type> u64_vals;
    vector<CString::type> strings;
    vector<bool> valid;
    vector<CDecimal *> decimals;
    
  public:
    CurrentValues(const int slots)
      :_slots(slots)
    {
      for (int i = 0; i < slots; ++i) {
        i64_vals.push_back(0);
        u64_vals.push_back(0);
        i32_vals.push_back(0);
        u32_vals.push_back(0);
        strings.push_back("");
        valid.push_back(false);
        decimals.push_back(0);
      }
    }

    
    virtual ~CurrentValues() {
      i32_vals.clear();
      u32_vals.clear();
      i64_vals.clear();
      u64_vals.clear();
      strings.clear();
      valid.clear();
      for (int i = 0; i < _slots; ++i) {
        if (decimals[i]) delete decimals[i];
        decimals[i] = 0;
      }
      decimals.clear();
    }


    vector<CInt32::type>& get_array_i32() { return i32_vals; }
    vector<CInt64::type>& get_array_i64() { return i64_vals; }
    vector<CuInt32::type>& get_array_u32() { return u32_vals; }
    vector<CuInt64::type>& get_array_u64() { return u64_vals; }
    vector<CString::type>& get_array_strings() { return strings; }
    vector<bool>& get_array_valid() { return valid; }
    vector<CDecimal *>& get_array_decimals() { return decimals; }
    

    bool is_valid(const int slot) const {
      assert(slot < (int) valid.size());
      return valid[slot];
    }

    
    void reset_valid(const int slot) {
      assert(slot < (int) valid.size());
      valid[slot] = false;
    }

    void set_valid(const int slot, const bool state) {
      assert(slot < (int) valid.size());
      valid[slot] = state;
    }

    CString::type get_string(const int slot) const {
      assert(slot < (int) strings.size());
      return strings[slot];
    }
    

    void set_string(const int slot, const CString::type& value) {
      assert(slot < (int) strings.size());
      strings[slot] = value;
      valid[slot] = true;
    }

    
    CuInt32::type get_u32(const int slot) const {
      assert(slot < (int) u32_vals.size());
      return u32_vals[slot];
    }

    
    CuInt32::type get(const int slot, const CuInt32& ignored) const {
      assert(slot < (int) u32_vals.size());
      return u32_vals[slot];
    }

    
    void set_u32(const int slot, const CuInt32::type value) {
      assert(slot < (int) u32_vals.size());
      u32_vals[slot] = value;
      valid[slot] = true;
    }

    
    CInt32::type get_i32(const int slot) const {
      assert(slot < (int) i32_vals.size());
      return i32_vals[slot];
    }
    

    CInt32::type get(const int slot, const CInt32& ignored) const {
      assert(slot < (int) i32_vals.size());
      return i32_vals[slot];
    }
    

    void set_i32(const int slot, const CInt32::type value) {
      assert(slot < (int) i32_vals.size());
      i32_vals[slot] = value;
      valid[slot] = true;
    }
    

    CuInt64::type get_u64(const int slot) const {
      assert(slot < (int) u64_vals.size());
      return u64_vals[slot];
    }
    

    CuInt64::type get(const int slot, const CuInt64& ignored) const {
      assert(slot < (int) u64_vals.size());
      return u64_vals[slot];
    }
    

    void set_u64(const int slot, const CuInt64::type value) {
      assert(slot < (int) u64_vals.size());
      u64_vals[slot] = value;
      valid[slot] = true;
    }

    
    CInt64::type get_i64(const int slot) const {
      assert(slot < (int) i64_vals.size());
      return i64_vals[slot];
    }

    
    CInt64::type get(const int slot, const CInt64& ignored) const {
      assert(slot < (int) i64_vals.size());
      return i64_vals[slot];
    }

    
    void set_i64(const int slot, const CInt64::type value) {
      assert(slot < (int) i64_vals.size());
      i64_vals[slot] = value;
      valid[slot] = true;
    }
    

    CDecimal* get_decimal(const int slot) const {
      assert(slot < (int) decimals.size());
      return decimals[slot];
    }
    

    void set_decimal(const int slot, CDecimal* value) {
      assert(slot < (int) decimals.size());
      if (!decimals[slot]) decimals[slot] = new CDecimal();
      decimals[slot]->set_exponent(value->get_exponent());
      decimals[slot]->set_mantissa(value->get_mantissa());
      valid[slot] = true;
    }
    

    void reset() {
      for (int i = 0; i < _slots; ++i) valid[i] = false;
    }
};


ostream& operator<<(ostream& out, const CurrentValues& cvs);

}

#endif /*CME_FAST_CURRENT_VALUES_H_*/
