// UMDF-FIX-Log decoder.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "LogReader.h"

#include <iostream>
#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/detail/lightweight_test.hpp>
#include <boost/shared_array.hpp>
#include <boost/asio.hpp>
#include <boost/lexical_cast.hpp>
#include <FASTDecoder/FASTDecoder.h>
#include <FASTDecoder/EventDecoder.h>
#include <FASTDecoder/FIXMessageEventHandler.h>

#pragma pack(push, 1)
struct MULTICAST_HEADER : boost::noncopyable
{
	unsigned long ulMsgSeqNum;
	unsigned short usNoChunks;
	unsigned short usCurrentChunk;
	unsigned short usMsgLength;
};
#pragma pack(pop)

boost::filesystem::ifstream input_file;
boost::filesystem::ofstream output_file;

void TBuffer::update(const unsigned char* pCurrentBuffer, std::size_t currentBufferSize)
{
	m_data.reset(new unsigned char[currentBufferSize]);
	memcpy(m_data.get(), pCurrentBuffer, currentBufferSize);
	m_dataSize = currentBufferSize;
}

void TBuffer::clear()
{
	m_data.reset();
	m_dataSize = 0;
}

CReceivedBuffer::CReceivedBuffer(unsigned short usNoChunks)
{
	m_bufferVector.reserve(usNoChunks);
	for (unsigned short i = 1; i <= usNoChunks; ++i)
	{
		m_usRemainingChunkSet.insert(i);
		m_bufferVector.push_back(TBuffer());
	}
}

size_t CReceivedBuffer::getRemainingChunk() const
{
	return m_usRemainingChunkSet.size();
}

size_t CReceivedBuffer::update(unsigned short usCurrentChunk, const unsigned char* pCurrentBuffer, std::size_t currentBufferSize)
{
	if(usCurrentChunk != 0)
	{
		TBuffer& buffer = m_bufferVector[usCurrentChunk - 1];
		buffer.update(pCurrentBuffer, currentBufferSize);

		m_usRemainingChunkSet.erase(usCurrentChunk);
		return getRemainingChunk();
	}
	else
		return 0;
}

boost::shared_ptr<TBuffer> CReceivedBuffer::getBufferPtr() const
{
	std::size_t i;
	std::size_t totalDataReceived = 0;

	for (i = 0; i < m_bufferVector.size(); ++i)
	{
		const TBuffer& buffer = m_bufferVector[i];
		totalDataReceived += buffer.getDataSize();
	}

	boost::shared_array<unsigned char> buffer(new unsigned char[totalDataReceived]);
	unsigned char* p = buffer.get();
	for (i = 0; i < m_bufferVector.size(); ++i)
	{
		const TBuffer& buffer = m_bufferVector[i];
		memcpy(p, buffer.getData().get(), buffer.getDataSize());
		p += buffer.getDataSize();
	}

	return boost::shared_ptr<TBuffer>(new TBuffer(buffer, totalDataReceived));
}

CMarketDataMessageEventHandler::CMarketDataMessageEventHandler() 
{}

CMarketDataMessageEventHandler::~CMarketDataMessageEventHandler(void)
{
}

// Event callbacks
void CMarketDataMessageEventHandler::onMessageBegin(int channelID, unsigned int sourceID)
{
	output_file << "8=FIX.4.4\001";
}

void CMarketDataMessageEventHandler::onMessageEnd(int channelID, unsigned int sourceID)
{
	output_file << "10=0\001" << std::endl;
}

void CMarketDataMessageEventHandler::onField(int fieldID, const std::string& value)
{
	output_file << fieldID << _T("=") << value << FAST::FieldDefinition::DELIMITER;
}

void CMarketDataMessageEventHandler::onRepeatingGroupBegin(int fieldID, unsigned int size)
{
	output_file << fieldID << _T("=") << size << FAST::FieldDefinition::DELIMITER;
}

void CMarketDataMessageEventHandler::onRepeatingGroupEnd(int fieldID)
{
}

void CMarketDataMessageEventHandler::onRepeatingGroupItemBegin(int slot)
{
}

void CMarketDataMessageEventHandler::onRepeatingGroupItemEnd(int slot)
{
}

void CMarketDataMessageEventHandler::onGroupBegin(int fieldID, int groupNo)
{
}

void CMarketDataMessageEventHandler::onGroupEnd(int fieldID)
{
}


int _tmain(int argc, _TCHAR* argv[])
{
	TMsgSeqNum2RecvBufferPtrMap m_msgSeqNum2RecvBufferPtrMap;

	boost::filesystem::path input_file_log_path( boost::filesystem::initial_path<boost::filesystem::path>() );
	boost::filesystem::path input_file_template_path( boost::filesystem::initial_path<boost::filesystem::path>() );
	boost::filesystem::path output_file_full_path( boost::filesystem::initial_path<boost::filesystem::path>() );

	if ( argc > 2 )
	{
		input_file_log_path = boost::filesystem::system_complete( boost::filesystem::path( argv[1] ) );
		input_file_template_path = boost::filesystem::system_complete( boost::filesystem::path( argv[2] ) );
		output_file_full_path = boost::filesystem::system_complete( boost::filesystem::path( argv[3] ) );

		if ( !boost::filesystem::exists( input_file_log_path ) )
		{
			std::cout << "\nNot found: " << input_file_log_path.file_string() << std::endl;
			return 1;
		}

		if ( !boost::filesystem::exists( input_file_template_path ) )
		{
			std::cout << "\nNot found: " << input_file_template_path.file_string() << std::endl;
			return 1;
		}

		if ( boost::filesystem::is_regular_file( input_file_log_path ) )
		{
			input_file.open( input_file_log_path.file_string() , std::ios_base::in | std::ios_base::binary);
			output_file.open( output_file_full_path.file_string() , std::ios_base::out );

			FAST::IFIXMessageEventHandlerPtr messageEventHandlerPtr(new CMarketDataMessageEventHandler());
			FAST::CEventDecoder decoder(input_file_template_path.file_string());
			decoder.registerEventHandler(messageEventHandlerPtr);
			
			while (!input_file.eof()) 
			{
				char* tech_header = new char[sizeof(MULTICAST_HEADER)];
				memset(tech_header, 0x00, sizeof(MULTICAST_HEADER));
				input_file.read (tech_header, sizeof(MULTICAST_HEADER));
				const MULTICAST_HEADER* pMulticastHeader = reinterpret_cast<const MULTICAST_HEADER*>(tech_header);
				unsigned long ulMsgSeqNum = ::ntohl(pMulticastHeader->ulMsgSeqNum);
				unsigned short usNoChunks = ::ntohs(pMulticastHeader->usNoChunks);
				unsigned short usCurrentChunk = ::ntohs(pMulticastHeader->usCurrentChunk);
				unsigned short usMsgLength = ::ntohs(pMulticastHeader->usMsgLength);
				
				char buffer [1024];
				memset(buffer, 0x00, 1024);

				output_file << std::endl << "[MsgSeqNum] [NoChunks] [CurrentChunk] [MsgLength]"<< std::endl;
				sprintf (buffer,"[%9d] [%8d] [%12d] [%9d] \n",ulMsgSeqNum, usNoChunks, usCurrentChunk, usMsgLength);
				output_file << buffer;
				output_file << "[Messages]"<< std::endl;
				
				char* fast_msg = new char[usMsgLength];
				memset(fast_msg, 0x00, usMsgLength);
				input_file.read (fast_msg, usMsgLength);

				boost::shared_ptr<CReceivedBuffer> receivedBufferPtr;

				TMsgSeqNum2RecvBufferPtrMap::iterator itMsgSeqNum = m_msgSeqNum2RecvBufferPtrMap.find(ulMsgSeqNum);
				if (itMsgSeqNum != m_msgSeqNum2RecvBufferPtrMap.end())
					receivedBufferPtr = itMsgSeqNum->second;
				else
					receivedBufferPtr.reset(new CReceivedBuffer(usNoChunks));

				size_t uiRemainingChunk = receivedBufferPtr->update(usCurrentChunk, reinterpret_cast<unsigned char*>(fast_msg) + sizeof(MULTICAST_HEADER), usMsgLength);
				if (!uiRemainingChunk)
				{
					boost::shared_ptr<TBuffer> bufferPtr = receivedBufferPtr->getBufferPtr();

					TDataJobPtr dataJobPtr(new CDataJob(bufferPtr->getData(), bufferPtr->getDataSize()));

					decoder.decode(reinterpret_cast<const char*>(dataJobPtr->getBuffer()), (int)dataJobPtr->getBufferSize());

					if (itMsgSeqNum != m_msgSeqNum2RecvBufferPtrMap.end())
						m_msgSeqNum2RecvBufferPtrMap.erase(itMsgSeqNum);
				}
				else
				{
					if (itMsgSeqNum == m_msgSeqNum2RecvBufferPtrMap.end())
						m_msgSeqNum2RecvBufferPtrMap.insert(TMsgSeqNum2RecvBufferPtrMap::value_type(ulMsgSeqNum, receivedBufferPtr));
				}
				delete [] fast_msg;
				fast_msg = NULL;
				delete [] tech_header;
				tech_header = NULL;
			}

			decoder.unregisterEventHandler(messageEventHandlerPtr);

			input_file.close();
			output_file.close();
		}
	}
	else
		std::cout << "\nUsage:  LogReader [path to UMDF log file] [path to FAST template file] [path to Report File]" << std::endl;

	return 0;
}

