#include "stdafx.h"
#include <string>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMNamedNodeMap.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/framework/LocalFileInputSource.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <xercesc/parsers/AbstractDOMParser.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMImplementationRegistry.hpp>
#include <xercesc/dom/DOMError.hpp>
#include <xercesc/dom/DOMLocator.hpp>
#include <xercesc/dom/DOMAttr.hpp>
#include <xercesc/dom/DOMErrorHandler.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>

#include "FASTDecoder.h"

using namespace xercesc;

namespace FAST
{

/**
 * This contains a pool of all the Element instances.  This is used to deallocate memory on Element destruction.
 */
long Element::g_ninstances = 0;

/**
 * This recursively builds an Element and children Elements.  All attributes are saved to a map.
 */
Element::Element(DOMElement* element, Element* parent = 0)
{
	assert(element != 0);
	this->parent = parent;

#ifdef _WIN32
	::InterlockedIncrement(&g_ninstances);
#else
	++g_ninstances;
#endif

	name = XMLString::transcode(element->getNodeName());

	// create all non-whitespace children:
	if (element->hasChildNodes())
	{
		DOMNodeList* sub_elements = element->getChildNodes();
		children.reserve(sub_elements->getLength());
		for (int i = 0, e = sub_elements->getLength(); i != e; ++i)
		{
            DOMNode* childNode = sub_elements->item(i);
            if (childNode->getNodeType() == DOMNode::ELEMENT_NODE) 
            {
			    DOMElement* child = (DOMElement *) childNode;
//			// ignore all #text elements.  This is the whitespace between elements in valid Template files.
//			if ("#text" == (string) XMLString::transcode(child->getNodeName()))
//				continue;
			    children.push_back(new Element(child, this));
            }
		}
	}
  
	// copy attributes into the attribute map:
	if (element->getAttributes())
	{
		DOMNamedNodeMap* attribs = element->getAttributes();
		for (int i = 0, e = attribs->getLength(); i != e; ++i)
		{
			DOMNode* attribute = attribs->item(i);

			string 
				key = XMLString::transcode( attribute->getNodeName()  ),
				value = XMLString::transcode( attribute->getNodeValue() );

			attributes[key] = value;
		}
	}
}


/**
 * Deallocates the memory created on the heap for this Element.
 * If there are no more instances in memory, the library is Terminated.
 */
Element::~Element()
{
	for (Iterator i = children.begin(), e = children.end(); i != e; ++i)
	{
		if (*i)
			delete *i;
		*i = NULL;
	}
	parent = NULL;
	children.clear();
	attributes.clear();

#ifdef _WIN32
	if (::InterlockedDecrement(&g_ninstances) <= 0)
		XMLPlatformUtils::Terminate();
#else
	if (--g_ninstances <= 0)
		XMLPlatformUtils::Terminate();
#endif  
}


Element* Element::get_root(const string& template_filename, bool validate)
	throw(CXMLParserException)
{
	if (g_ninstances <= 0)
		XMLPlatformUtils::Initialize();

	// Instantiate the DOM parser.
	bool bNamespaces = true;
	bool bSchema = true;
	bool bSchemaFullChecking = true;
	bool bDOMValidateIfSchema = true;

	const char* xmlFile = template_filename.c_str();
	XercesDOMParser *parser = new XercesDOMParser();
	if (validate == true)
	{
        parser->setDoNamespaces (true);
        parser->setDoSchema (true);
        parser->setValidationSchemaFullChecking (true);
	}


	// Create our error handler and install it
	CXMLDebugErrorHandler errorHandler;
	parser->setErrorHandler(&errorHandler);

    //reset error count first
    errorHandler.resetErrors();

	// reset document pool
    parser->resetDocumentPool();

	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *document = NULL;
	try
	{
		parser->parse(xmlFile);
		document = parser->getDocument();
	}
	catch (const XMLException& toCatch)
	{
		StrX strX(toCatch.getMessage());
		throw CXMLParserException(strX.localForm());
	}
	catch (const DOMException& toCatch)
	{
        const unsigned int maxChars = 2047;
        XMLCh errText[maxChars + 1];
		stringstream ss;
		ss << _T("DOM Error during parsing: [") << xmlFile << _T("] - ");
		ss << _T("DOM Exception code=[") << toCatch.code << _T("]");
        if (DOMImplementation::loadDOMExceptionMsg(toCatch.code, errText, maxChars))
			ss << _T(", Message=[") << StrX(errText) << _T("]");
		throw CXMLParserException(ss.str().c_str());
	}
	catch (...)
	{
		stringstream ss;
		ss << _T("Unexpected exception during parsing: [") << xmlFile << _T("]");
		throw CXMLParserException(ss.str().c_str());
	}


	if (document == NULL)
		return NULL;

	return new Element(document->getDocumentElement());
}

Element* Element::get_root_from_buffer(const XMLByte* const buffer, XMLSize_t bufferSz, bool validate)
	throw(CXMLParserException)
{
	if (g_ninstances <= 0)
		XMLPlatformUtils::Initialize();

	// Generate MemBufInputSource
	XERCES_CPP_NAMESPACE::MemBufInputSource memBuffer(buffer, bufferSz, L"template_stream");

	// Instantiate the DOM parser.
	bool bNamespaces = true;
	bool bSchema = true;
	bool bSchemaFullChecking = true;
	bool bDOMValidateIfSchema = true;

	XercesDOMParser *parser = new XercesDOMParser();
	if (validate == true)
	{
        parser->setDoNamespaces(true);
        parser->setDoSchema(true);
        parser->setValidationSchemaFullChecking(true);
	}


	// Create our error handler and install it
	CXMLDebugErrorHandler errorHandler;
	parser->setErrorHandler(&errorHandler);

    //reset error count first
    errorHandler.resetErrors();

	// reset document pool
    parser->resetDocumentPool();

	XERCES_CPP_NAMESPACE_QUALIFIER DOMDocument *document = NULL;
	try
	{
		parser->parse(memBuffer);
		document = parser->getDocument();
	}
	catch (const XMLException& toCatch)
	{
		StrX strX(toCatch.getMessage());
		throw CXMLParserException(strX.localForm());
	}
	catch (const DOMException& toCatch)
	{
        const unsigned int maxChars = 2047;
        XMLCh errText[maxChars + 1];
		stringstream ss;
		ss << _T("DOM Error during parsing buffer = [") << buffer << _T("]");
		ss << _T("DOM Exception code=[") << toCatch.code << _T("]");
        if (DOMImplementation::loadDOMExceptionMsg(toCatch.code, errText, maxChars))
			ss << _T(", Message=[") << StrX(errText) << _T("]");
		throw CXMLParserException(ss.str().c_str());
	}
	catch (...)
	{
		stringstream ss;
		ss << _T("Unexpected exception during parsing buffer = [") << buffer << _T("]");
		throw CXMLParserException(ss.str().c_str());
	}


	if (document == NULL)
		return NULL;

	return new Element(document->getDocumentElement());
}

}