#pragma once

#include <windows.h>
#include <mmsystem.h>
#include <iostream>
#include <string>
#include <vector>
#include <cstdio>

/// <summary>Generates a nanosecond timestamp, according to PUMA requirements
/// (number of milliseconds since the Unix Epoch). 
/// </summary>
class NanosTimestamp {
public:
    /// <summary>Constructor</summary>
    NanosTimestamp() { 
        ::QueryPerformanceFrequency (&ticksPerSecond);
        if (previousTime == 0) {
            ::GetSystemTimeAsFileTime (&ftSystemTime);
            previousTime = ft2ns (ftSystemTime);
            ::QueryPerformanceCounter (&reference);
        }
    }
    /// <summary>Tries to return a nanosecond timestamp, according to PUMA requirements</summary>
    /// <returns>The nanosecond timestamp</returns>
    __int64 getTimestamp() {
        // Compare the value returned by the API and checks if it is equal to the previous value.
        // If it is equal, offset it using the QueryPerformanceCounter / QueryPerformanceFrequency combination.
        LARGE_INTEGER now;
        ::QueryPerformanceCounter (&now);
        return previousTime + diff2ns (now, reference, ticksPerSecond); // adds the difference measured by QueryPerformanceCounter
    }
    /// <summary>Tries to return a nanosecond timestamp, according to PUMA requirements</summary>
    /// <param name='timestamp'>The timestamp, formatted as YYYYMMDDHHMMSSmmm - example: 20101222132142823</param>
    /// <returns>The nanosecond timestamp</returns>
    __int64 getTimestamp(std::string& timestamp) {
        FILETIME ftNow;
        SYSTEMTIME stNow;
        // Compare the value returned by the API and checks if it is equal to the previous value.
        // If it is equal, offset it using the QueryPerformanceCounter / QueryPerformanceFrequency combination.
        LARGE_INTEGER now;
        ::QueryPerformanceCounter (&now);
        char szBuff[32];
        __int64 diff = diff2ns (now, reference, ticksPerSecond); // adds the difference measured by QueryPerformanceCounter
        ftadd (ftNow, ftSystemTime, diff);
        ::FileTimeToSystemTime (&ftNow, &stNow);
        sprintf_s (szBuff, sizeof(szBuff), "%04d%02d%02d%02d%02d%02d%03d", stNow.wYear, stNow.wMonth, stNow.wDay, 
            stNow.wHour, stNow.wMinute, stNow.wSecond, stNow.wMilliseconds);
        timestamp.assign (szBuff);
        return previousTime + diff; 
    }
private:
    /// <summary>Converts a FILETIME into nanoseconds since 1/1/1970 (UTC)</summary>
    static __int64 ft2ns (const FILETIME& ft) {
        ULARGE_INTEGER uli;
        uli.LowPart = ft.dwLowDateTime;
        uli.HighPart = ft.dwHighDateTime;
        // FILETIME is the number of 100-nanosecond intervals since January 1, 1601 (UTC)
        // So it must be normalized to the number of nanoseconds since January 1, 1970 (UTC)
        // difference between 1/1/1601 and 1/1/1970 in seconds is 11644473600
        return 100LL * (uli.QuadPart - 116444736000000000LL);
    }
    /// <summary>Adds nanoseconds to a FILETIME</summary>
    /// <param name='ftDest'>ftSrc + ns</param>
    /// <param name='ftSrc'>The source FILETIME</param>
    /// <param name='ns'>The difference time in nanoseconds - it will be truncated to a 100 ns multiple</param>
    static void ftadd (FILETIME& ftDest, const FILETIME& ftSrc, __int64 ns) {
        ULARGE_INTEGER uiTime;
        uiTime.LowPart  = ftSrc.dwLowDateTime;
        uiTime.HighPart = ftSrc.dwHighDateTime;
        uiTime.QuadPart += ns / 100;
        ftDest.dwLowDateTime  = uiTime.LowPart;
        ftDest.dwHighDateTime = uiTime.HighPart;
    }
public:    
    /// <summary>Returns the difference between two values taken by QueryPerformanceCounter, in nanoseconds</summary>
    static __int64 diff2ns (const LARGE_INTEGER& now, const LARGE_INTEGER& reference, const LARGE_INTEGER& ticksPerSecond) {
        __int64 diff = (now.QuadPart - reference.QuadPart);
        if (diff <= 9223372035L)
            return diff * 1000000000LL / ticksPerSecond.QuadPart;
        else if (diff <= 92233720350LL)
            return (diff * 100000000LL / (ticksPerSecond.QuadPart / 10LL));
        else if (diff <= 922337203500L)
            return (diff * 10000000LL / (ticksPerSecond.QuadPart / 100LL));
        else if (diff <= 9223372035000LL)
            return (diff * 1000000LL / (ticksPerSecond.QuadPart / 1000LL));
        else if (diff <= 92233720350000LL)
            return (diff * 100000LL / (ticksPerSecond.QuadPart / 10000LL));
        else if (diff <= 922337203500000LL)
            return (diff * 10000LL / (ticksPerSecond.QuadPart / 100000LL));
        else if (diff <= 9223372035000000LL)
            return (diff * 1000LL / (ticksPerSecond.QuadPart / 1000000LL));
        else
            return (diff * 100LL / (ticksPerSecond.QuadPart / 10000000LL));
    }
private:
    static __int64 previousTime;
    static LARGE_INTEGER reference;
    static FILETIME ftSystemTime;
    LARGE_INTEGER ticksPerSecond;
};

