// FastLogger.cpp : Defines the entry point for the console application.
//
#include "stdafx.h"

#include "UMDFMarketDataReceiver.h"
#include "MarketDataReceiverConfiguration.h"

std::string _dumpMessagesPath;
std::string _readDumpPath;
bool _dumpMessages;
int _dumpSources[4];

using namespace BmfMessagingPlugin;

class UMDFLogger {
private:
	std::string _pluginName;
	std::string _pluginChannel;
	std::string _configurationFile;
	std::string _dumpPath;
	std::string _readDump;
	HANDLE _stopServiceEvent;
	
public:
	UMDFLogger(LPCTSTR name) :
	  _stopServiceEvent(NULL)
	{
		_pluginName = name;

		BOOL bInitialState = FALSE;
		BOOL bManualReset = TRUE;
		LPSECURITY_ATTRIBUTES lpEventAttributes = NULL;
		LPTSTR lpName = NULL;
		_stopServiceEvent = ::CreateEvent(lpEventAttributes, bManualReset, bInitialState, lpName);
		ASSERT(NULL != _stopServiceEvent);

		_dumpSources[0] = 0; _dumpSources[1] = 0; _dumpSources[2] = 0; _dumpSources[3] = 0;
	}

	virtual ~UMDFLogger() {}

	void usage(const char *name) {

		printf("usage of %s:\r\n", name);
		printf("-ch           Channel to receive\r\n");
		printf("-src          Sources to receive [optional]\r\n");
		printf("-cfg/-conf/-c Config file\r\n");
		printf("-o/-out       Output directory\r\n");
		printf("-read         Read dump files in set directory [optional]\r\n");
		printf("\r\nExamples:\r\n");
		printf("%s -ch 8 -c marketDataReceiver.conf -o d:\\logs \r\n", name);
		printf("%s -ch 8 -c marketDataReceiver.conf -read d:\\dump -o d:\\logs \r\n", name);
		printf("%s -src 0,1 -ch 8 -c marketDataReceiver.conf -o d:\\logs \r\n", name);

	}

	bool load(int argc, _TCHAR* argv[])
	{
		bool hasSrc = false;
		for(int i=1; i < argc; i++) {
			const char *arg = argv[i];

			if( stricmp(arg, "-ch") == 0 ) {
				_pluginChannel = argv[i+1];
				i++;
				continue;
			}

			if( stricmp(arg, "-src") == 0 ) {
				const char *p = argv[i+1];

				if( strstr(p, "0") != NULL ) _dumpSources[0] = 1;
				if( strstr(p, "1") != NULL ) _dumpSources[1] = 1;
				if( strstr(p, "3") != NULL ) _dumpSources[3] = 1;

				if( strstr(p, "0") != NULL ) printf("FONTE 0\r\n");
				if( strstr(p, "1") != NULL ) printf("FONTE 1\r\n");
				if( strstr(p, "3") != NULL ) printf("FONTE 3\r\n");

				hasSrc = true;
				
				i++;
				continue;
			}

			if( stricmp(arg, "-cfg") == 0 || stricmp(arg, "-conf") == 0 || stricmp(arg, "-c") == 0 ) {
				_configurationFile = argv[i+1];
				i++;
				continue;
			}

			if( stricmp(arg, "-o") == 0 || stricmp(arg, "-out") == 0 ) {
				_dumpPath = argv[i+1];
				i++;
				continue;
			}

			if( stricmp(arg, "-read") == 0 ) {
				_readDump = argv[i+1];
				i++;
				continue;
			}

		}

		if( _pluginChannel.empty() || _configurationFile.empty() || _dumpPath.empty() ) {
			usage(argv[0]);
			return false;
		}

		if( !hasSrc ) {
			_dumpSources[0] = 1;
			_dumpSources[1] = 1;
			_dumpSources[2] = 1;
			_dumpSources[3] = 1;
		}

		if( _dumpPath.size() > 0 ) {
			_dumpMessages = true;
			_dumpMessagesPath = _dumpPath;
		}

		if( _readDump.size() > 0 ) {
			_readDumpPath = _readDump;
		}

		return true;
	}

	bool run() {

		BOOL bResult = FALSE;
		MarketDataReceiverConfiguration* pConfiguration = NULL;

		try
		{
			MarketDataReceiverConfiguration::RegisterConfigurationInfo(GetConfigurationFile());
			//MarketDataReceiverConfiguration::RegisterPluginChannel(GetPluginChannel());
			pConfiguration = &MarketDataReceiverConfiguration::getInstance();
			ASSERT(pConfiguration != NULL);
			if (pConfiguration == NULL)
				throw std::runtime_error(_T("BTS UMDF Market Data Receiver configuration failed!"));

			TCHAR szFilePath[_MAX_PATH];
			DWORD dwRet = ::GetModuleFileName(NULL, szFilePath, _MAX_PATH);
			if (!dwRet) {
				std::stringstream ss;
				ss << "BTS UMDF Market Data Receiver Startup" << ::GetLastError();
				throw std::runtime_error( ss.str() );
			}
			//GW_WARNING_FORMAT(_T("** Configuration Loaded from %s"), pConfiguration->getAppLogFilePath())

			// Initialize receiver instances
			TUMDFMarketDataReceiverPtrMap marketDataReceiverPtrMap;

			for (TChannelConfigurationPtrVector::const_iterator itChannel = pConfiguration->getChannelConfigurationPtrVector().begin(); itChannel != pConfiguration->getChannelConfigurationPtrVector().end(); ++itChannel)
			{
				TChannelConfigurationPtr channelConfigurationPtr = *itChannel;
				//TODO: passes the plugin channel to configuration loader class.
				if (channelConfigurationPtr->getNumber() == _ttoi(GetPluginChannel()))
				{
					GW_INFO_FORMAT("Creating UMDFMarketDataReceiver[%d]: %s ...", channelConfigurationPtr->getNumber() % channelConfigurationPtr->getDefinition().c_str());
					TUMDFMarketDataReceiverPtr marketDataReceiverPtr(new CUMDFMarketDataReceiver(channelConfigurationPtr));
					marketDataReceiverPtrMap.insert(TUMDFMarketDataReceiverPtrMap::value_type(channelConfigurationPtr->getNumber(), marketDataReceiverPtr));
					GW_INFO_FORMAT(_T("Starting UMDFMarketDataReceiver[%d]: %s ..."), channelConfigurationPtr->getNumber() % channelConfigurationPtr->getDefinition().c_str())
					marketDataReceiverPtr->start(getStopServiceEvent());
					break;
				}
			}

			//LPRINT(_T("UMDFMarketDataReceiver started!"))
			//WaitForSingleObject(getStopServiceEvent(), INFINITE);
			printf("\r\nPress any key to exit...\r\n");
			getchar();
			OnStop();

		}
		catch (std::exception &err) 
		{
			// It should also catch all FIX::Exception at startup
			GW_ERROR_FORMAT("Error loading UMDFLogger - ", err.what());
			halt();
			return false;
		}
		catch (...)
		{
			GW_ERROR("Error when loading UMDFLogger");
			halt();
			return false;
		}

		return true;
	}
	void OnStop() {
		SetEvent(_stopServiceEvent);
	}
	
	LPCTSTR GetPluginID() const 
	{
		return _pluginName.c_str();
	}

	LPCTSTR GetPluginChannel() const 
	{
		return _pluginChannel.c_str();
	}

	LPCTSTR GetConfigurationFile() const 
	{
		return _configurationFile.c_str();
	}

	HANDLE getStopServiceEvent() 
	{
		return _stopServiceEvent;
	}

	void halt() {
		/*if (m_pGatewayApp != NULL)
			m_pGatewayApp->HaltService(this);
		*/
		exit(-1);
	}

};

int _tmain(int argc, _TCHAR* argv[])
{
	printf("UMDFLogger starting...\r\n");

	UMDFLogger logger("UMDF LOGGER");
	if( !logger.load(argc, argv) ) {
		printf("\r\nPress any key to exit...\r\n");
		getchar();
		return -1;
	}

	GW_INFO("UMDFLogger starting...");
	if (logger.run()) {
		printf("UMDFLogger stopped !!\r\n");
		GW_INFO("UMDFLogger stopped !!");
	}

	Sleep(5000);
	//getchar();

	return 0;
}

