#include "StdAfx.h"
#include <vector>
#include <sstream>

#include "FASTDecoder.h"
#include "EventDecoder.h"

#ifdef _DUMP_MESSAGES
CRITICAL_SECTION _cs;
static bool _init=false;
#endif

namespace FAST
{

const string CEventDecoder::VERSION = "1.0.1.1";


/**
 * @return a string describing the version number.
 */
string CEventDecoder::get_version()
{
  stringstream ss;
  ss << "BM&F C++ FixFAST Decoder v" << VERSION;
  return ss.str();
}


CEventDecoder::CEventDecoder(void) : m_codec(NULL)
{
#ifdef _DUMP_MESSAGES
	if( !_init ) { InitializeCriticalSection(&_cs); _init = true; }
#endif
}

CEventDecoder::CEventDecoder(const string& template_file) : m_codec(NULL)
{
  this->m_codec = this->m_templateBuilder.create_templates(template_file, this->m_templates, &this->m_transferEncoder);
#ifdef _DUMP_MESSAGES
  if( !_init ) { InitializeCriticalSection(&_cs); _init = true; }
#endif
}

CEventDecoder::CEventDecoder(const unsigned char* const buffer, unsigned long bufferSz) : m_codec(NULL)
{
  this->m_codec = this->m_templateBuilder.create_templates(buffer, bufferSz, this->m_templates, &this->m_transferEncoder);
#ifdef _DUMP_MESSAGES
  if( !_init ) { InitializeCriticalSection(&_cs); _init = true; }
#endif
}

CEventDecoder::~CEventDecoder(void)
{
	this->m_codec = NULL;
}

// Register/Unregister
bool CEventDecoder::registerEventHandler(IFIXMessageEventHandlerPtr fixMessageEventHandlerPtr)
{
	return this->m_eventHandlerContainer.registerEventHandler(fixMessageEventHandlerPtr);
}

bool CEventDecoder::unregisterEventHandler(IFIXMessageEventHandlerPtr fixMessageEventHandlerPtr)
{
	return this->m_eventHandlerContainer.unregisterEventHandler(fixMessageEventHandlerPtr);
}


// Start de decoding process
void CEventDecoder::decode(const char* msg, int bufferSize)
{
	this->decode(0, 0, msg, bufferSize);
}

void CEventDecoder::decode(int channelID, unsigned int sourceID, const char* msg, int bufferSize)
{
#ifdef _DUMP_MESSAGES
	EnterCriticalSection(&_cs);
#endif

    assert(this->m_codec != NULL);

	const CBytes::byte* buffer = reinterpret_cast<const CBytes::byte *>(msg);
	this->m_transferEncoder.set_input_stream(buffer, bufferSize);
	this->m_transferEncoder.reset_input_pointer();

	std::vector<Codec *> used;
	CBytesPtr pmap = this->m_codec->get_presence_map_off_stream();

	while (pmap->get_length() > 0)
	{
		this->m_codec->parse_pmap(pmap, true, 0, true);

		// Parse the template id
		U32FieldDefinition* template_field = reinterpret_cast<U32FieldDefinition *>(this->m_codec->get_definitions()[0]);

		template_field->decode(0, *this->m_codec);
		string templateId = template_field->get_value();

		if (Decoder::LOGME)
		{
			FLogger::get_instance().append_buffer("Received templateID: ");
			FLogger::get_instance().append_buffer(templateId.c_str());
			FLogger::get_instance().print_buffer();
			FLogger::get_instance().reset_buffer();
		}

		// Switch Templates
		if (this->m_codec->get_template_id() != (reinterpret_cast<CuInt32 *>(template_field->get_data()))->get_value())
		{
			TemplateBuilder::TemplateMap::const_iterator it_template = this->m_templates.find(templateId);
			if (it_template != this->m_templates.end())
			{
				this->m_codec = it_template->second;
				if (Decoder::LOGME)
					FLogger::get_instance().reset_buffer();
			}
		}

		this->m_codec->parse_pmap(pmap, true, 0, false);

		// Log the presence map & template id
		if (Decoder::LOGME)
		{
			FLogger::get_instance().print_buffer();
			FLogger::get_instance().reset_buffer();
			FLogger::get_instance().print_header();
			FLogger::get_instance().pretty_print_display();
			FLogger::get_instance().reset();
		}

		vector<FieldDefinition *>& fieldDefinitions = this->m_codec->get_definitions();
		// Pre-condition: every stream contains only one Fix message
		this->m_eventHandlerContainer.onMessageBegin(channelID, sourceID);
		for (unsigned int i=1; i<fieldDefinitions.size(); ++i)
			fieldDefinitions[i]->decode(i, *this->m_codec, this->m_eventHandlerContainer);
		// Pre-condition: every stream contains only one Fix message
		this->m_eventHandlerContainer.onMessageEnd(channelID, sourceID);

		if (Decoder::LOGME)
			FLogger::get_instance().println();

		used.push_back(this->m_codec);
		pmap = this->m_codec->get_presence_map_off_stream();
	}

	// Clear dictionary after a batch.
	for (vector<Codec *>::iterator i = used.begin(), e = used.end(); i != e; ++i)
		(*i)->reset_dictionary();
	used.clear();

	this->m_transferEncoder.reset_input_pointer();

#ifdef _DUMP_MESSAGES
	LeaveCriticalSection(&_cs);
#endif
	
}

}