#pragma once
#ifndef __CMARKETDATARECEIVERCONFIGURATION_H__
#define __CMARKETDATARECEIVERCONFIGURATION_H__

#include "stdafx.h"
#include <Element.h>

namespace BmfMessagingPlugin
{

	class CChannelConfigurationException : public std::logic_error
	{
	public:
		CChannelConfigurationException(const std::string& what_) : std::logic_error(what_) {}
		virtual ~CChannelConfigurationException(void) {}
	};

	class  CSourceConfiguration : private boost::noncopyable
	{
	public:
		enum EnumSourceType
		{
			InstrumentReplay,
			Snapshot,
			HistoricalReplay,
			Incremental
		};

		CSourceConfiguration(EnumSourceType sourceType, const std::string& localAddr, const std::string& remoteAddr, int remotePort, int receiveBufferSize, std::size_t noWorkerThreads);
		CSourceConfiguration(EnumSourceType sourceType, const std::string& transportId, const std::string& topicId, const std::string& receiverId);
		CSourceConfiguration(FAST::Element* pSourceConfigurationElement, std::size_t noWorkerThreads);
		virtual ~CSourceConfiguration();

		EnumSourceType getSourceType() const { return this->m_sourceType; }
		bool isUsingCaml() const { return this->m_useCaml; }
		const std::string& getLocalAddr() const { return this->m_localAddr; }
		const std::string& getRemoteAddr() const { return this->m_remoteAddr; }
		int getRemotePort() const { return this->m_remotePort; }
		int getReceiveBufferSize() const { return this->m_receiveBufferSize; }
		int getChunkFastTimeoutSeconds() const { return this->m_chunkFastTimeoutSeconds; }
		std::size_t getNoWorkerThreads() const { return this->m_noWorkerThreads; }
		const std::string& getUsername() const { return this->m_username; }
		const std::string& getPassword() const { return this->m_password; }
		unsigned int getTimeout() const { return this->m_timeout; }

		const std::string& getTransportId() const { return this->m_transportId; }
		const std::string& getTopicId() const { return this->m_topicId; }
		const std::string& getReceiverId() const { return this->m_receiverId; }

	private:
		EnumSourceType m_sourceType;
		bool m_useCaml;
		std::string m_localAddr;
		std::string m_remoteAddr;
		int m_remotePort;
		int m_receiveBufferSize;
		int m_chunkFastTimeoutSeconds;
		std::size_t m_noWorkerThreads;

		// Used only for SourceType = HistoricalReplay (TCP replay)
		std::string m_username;
		std::string m_password;
		unsigned int m_timeout;

		// Used only when Incremental feed is send thru tibco/caml
		std::string m_transportId;
		std::string m_topicId;
		std::string m_receiverId;
	};

	typedef boost::shared_ptr<CSourceConfiguration> TSourceConfigurationPtr;

	struct TSourceConfigurationPtrLess : public std::binary_function<TSourceConfigurationPtr, TSourceConfigurationPtr, bool>
	{
		bool operator()(TSourceConfigurationPtr confPtr1, TSourceConfigurationPtr confPtr2) const
		{
			return (confPtr1->getSourceType() < confPtr2->getSourceType());
		}
	};

	typedef std::map<CSourceConfiguration::EnumSourceType, TSourceConfigurationPtr> TSourceConfigurationPtrMap;

	class  CChannelConfiguration : private boost::noncopyable
	{
	public:
		static const size_t MAX_NOMSG_QUEUE = 99999;

		CChannelConfiguration(FAST::Element* pChannelConfigurationElement, 
							  const std::string& templateFilePath, 
							  const std::string& camlConfigFilePath);
		virtual ~CChannelConfiguration(void);
		int getNumber() const { return this->m_number; }
		const std::string& getDefinition() const { return this->m_definition; }
		const std::string& getDefaultSecurityExchange() const { return this->m_defaultSecurityExchange; }
		const std::string& getSecurityPrefix() const { return this->m_securityPrefix; }
		const std::string& getTemplateFilePath() const { return this->m_templateFilePath; }
		const std::string& getCamlConfigFilePath() const { return this->m_camlConfigFilePath; }
		int getSmallScaleRetries() const { return this->m_smallScaleRetries; }
		int getLargeScaleRetries() const { return this->m_largeScaleRetries; }
		long getMaxNoMessageQueue() const { return this->m_maxNoMessageQueue; }
		DWORD getRetryTimeout() const { return this->m_retryTimeout; }
		const TSourceConfigurationPtrMap& getSourceConfigurationPtrMap() const { return this->m_sourceConfigurationPtrMap; }
		TSourceConfigurationPtr getSourceConfigurationPtr(CSourceConfiguration::EnumSourceType sourceType) const;

	private:
		int m_number;
		std::string m_definition;
		std::string m_defaultSecurityExchange;
		std::string m_securityPrefix;
		int m_smallScaleRetries;
		int m_largeScaleRetries;
		long m_maxNoMessageQueue;
		DWORD m_retryTimeout;

		std::string m_templateFilePath;
		std::string m_camlConfigFilePath;
		TSourceConfigurationPtrMap m_sourceConfigurationPtrMap;
	};

	typedef boost::shared_ptr<CChannelConfiguration> TChannelConfigurationPtr;
	typedef std::vector<TChannelConfigurationPtr> TChannelConfigurationPtrVector;

	class  MarketDataReceiverConfiguration : private boost::noncopyable
	{
	public:

		const TChannelConfigurationPtrVector& getChannelConfigurationPtrVector() const { return this->m_channelConfigurationPtrVector; }

		static MarketDataReceiverConfiguration &getInstance() {
			static MarketDataReceiverConfiguration *inst = NULL;
			if( inst == NULL ) {
				inst = CreateInstance();
			}
			return *inst;
		}

		inline static void RegisterConfigurationInfo(LPCTSTR configurationFilePath)
		{ 
			g_configurationFilePath = configurationFilePath;
		}
		inline static void RegisterPluginChannel(LPCTSTR pluginChannel)
		{ 
			g_pluginChannel = pluginChannel;
		}

		void parseConfiguration();

	protected:
		static MarketDataReceiverConfiguration* CreateInstance()
		{
			assert(!g_configurationFilePath.empty());
			MarketDataReceiverConfiguration *conf = new MarketDataReceiverConfiguration();
			conf->parseConfiguration();
			return conf;
		}

	private:
		static std::string g_configurationFilePath;
		static std::string g_pluginChannel;
		TChannelConfigurationPtrVector m_channelConfigurationPtrVector;

	};

	//typedef BmfStd::CSingleton<CMarketDataReceiverConfiguration, MarketDataReceiverConfiguration, BmfWin32::CCriticalSection> CMarketDataReceiverConfigurationSingleton;

}

#endif	// __CMARKETDATARECEIVERCONFIGURATION_H__
